@@ -12,6 +12,7 @@ use std::time::Duration;
 use tokio::sync::RwLock;
 
 pub mod actions;
+pub mod helpers;
 mod iteration;
 mod result;
 
@@ -42,30 +43,25 @@ pub struct StaticContext {
 
 #[derive(Debug, Error)]
 pub enum StaticContextError {
-    #[error("templating error: `{0}`")]
+    #[error("templating error: {0}")]
     TemplateError(#[from] liquid::Error),
-    #[error("error deserializing properties: `{0}`")]
+    #[error("error deserializing properties: {0}")]
     YamlDeserializationError(#[from] serde_yaml::Error),
+    #[error("error deserializing json: {0}")]
+    JsonDeserialization(#[from] serde_json::Error),
+    #[error("io error: {0}")]
+    IoError(#[from] std::io::Error),
+    #[error("csv error: {0}")]
+    CsvError(#[from] csv::Error),
 }
 #[derive(Debug, Error)]
 pub enum StaticContextActionBuilderError {
-    #[error("error deserializing properties: `{0}`")]
+    #[error("error deserializing properties: {0}")]
     YamlDeserializationError(#[from] serde_yaml::Error),
-    #[error("error building action `{0}`: `{1}`")]
+    #[error("error building action {0}: {1}")]
     ValidationError(&'static str, String),
 }
 
-/// A helper to build an error when field validation failed on a context action builder
-pub fn context_validation_error<S: Into<String>>(
-    builder: &impl StaticContextActionBuilder,
-    msg: S,
-) -> Result<StaticContextActionPointer, StaticContextActionBuilderError> {
-    Err(StaticContextActionBuilderError::ValidationError(
-        builder.get_type_name(),
-        msg.into(),
-    ))
-}
-
 #[derive(Debug)]
 pub enum IterationContextItem {}
 
@@ -74,15 +70,18 @@ pub struct IterationContext {
     pub items: ContextMap,
 }
 
+#[derive(Debug, Error)]
+pub enum IterationContextError {
+    #[error("error from static context {0}: {1}")]
+    StaticContext(String, StaticContextError),
+}
+
 #[derive(Debug)]
 pub(crate) enum IterationResult {
     Ok {
         actions: Vec<InternalActionResult>,
         context: IterationContext,
     },
-    ContextError {
-        err: StaticContextError,
-    },
 }
 
 #[derive(Serialize, Debug, Default, Clone)]
@@ -168,18 +167,37 @@ impl TestRunContext {
         }
     }
 
-    pub(crate) async fn new_iteration_ctx(&self) -> Result<IterationContext, StaticContextError> {
+    pub(crate) async fn new_iteration_ctx(
+        &self,
+    ) -> Result<IterationContext, IterationContextError> {
         let mut items = HashMap::with_capacity(self.static_contexts.len());
         for v in &self.static_contexts {
-            let v = v.value().get_val().await?;
-            for (k, v) in v.items {
-                items.insert(k.clone(), to_value(&v)?);
+            let name = v.key().clone();
+            if let Err(e) = self.write_static_context(&mut items, name.as_str()).await {
+                return Err(IterationContextError::StaticContext(name, e));
             }
         }
 
         Ok(IterationContext { items })
     }
 
+    async fn write_static_context(
+        &self,
+        items: &mut HashMap<String, Value>,
+        key: &str,
+    ) -> Result<(), StaticContextError> {
+        let v = self
+            .static_contexts
+            .get(key)
+            .expect("should be called from an iterator");
+        let v = v.value().get_val().await?;
+        for (k, v) in v.items {
+            items.insert(k.clone(), to_value(&v)?);
+        }
+
+        Ok(())
+    }
+
     pub(crate) fn is_not_done(&self) -> bool {
         !self.duration_tracker.is_done()
     }
